import win32.windows;
import win32.windef;
import win32.commctrl;
import std.stdio;
import std.string;
import std.utf;
import std.math;

import GraphicBaseEx;

T SGN(T)(T a) { return( a<0 ? -1 : a>0 ? 1 : 0 ) ;}
T ABS(T)(T a) { return( a<0 ? -a : a ) ;}

enum MOUSE_OPSTATE{
  IDLE,
  NONE,
  CTRL
}

class GraphicPanel : GraphicBaseEx {
  int    prev_mouse_Button ;
  int    start_x ;
  int    start_y ;
  int    scale_center_x ;
  int    scale_center_y ;
  int    dist_y ;
  int    viewpos_x ;
  int    viewpos_y ;
  double prev_scale ;
  const double maxscale=128.0 ;
  const double minscale=0.001 ;
  MOUSE_OPSTATE opstat=MOUSE_OPSTATE.IDLE ;

  this(string name, string ClassName) {
    super(name,ClassName) ;
  }

  override int wm_create(){
    return(0) ;
  }

  override int wm_paint(){
    return(0) ;
  }

  override int wm_size(int w, int h){
    return(0) ;
  }

  override int wm_mousewheel(int d, int key){
    if( d!=0 ){
      double scale=cast(double)(d)/120.0 ;
      double dispscale= getScale() * (1.0 + (getZoomrate() * SGN(scale))) ;
      scaling(dispscale) ;
      return(1) ;
    }
    return(0) ;
  }

  override int mouse( int acs_stat, int mb, int mx, int my ){
    if( (mb & 0x10)!=0 || (mb & 0x20)!=0 ){
      if( (mb & 0x10)!=0 ) sendmsgtoparent(WM_RBUTTONDOWN, 0,0) ;
      if( (mb & 0x20)!=0 ) sendmsgtoparent(WM_RBUTTONUP,   0,0) ;
    }
    if( (acs_stat & CNTL_KEY)!=0 ){
      if( (prev_mouse_Button & 1)==0 && (mb & 1)!=0 ){  // L-button off => on
        prev_scale=getScale() ;
        getRawMouseCurPos(&start_x, &start_y) ;
        dist_y=0 ;
        scale_center_x=winW/2 ;
        scale_center_y=winH/2 ;
        prev_mouse_Button=mb ;
        opstat=MOUSE_OPSTATE.CTRL ;
        version(MOUSE_OP_TUNED){
          ShowCursor(false) ;
        }
      }
    }else{
      if( (prev_mouse_Button & 1)==0 && (mb & 1)!=0 ){  // L-button off => on
        getRawMouseCurPos(&start_x, &start_y) ;
        getViewPosition(&viewpos_x, &viewpos_y) ;
        prev_mouse_Button=mb ;
        opstat=MOUSE_OPSTATE.NONE ;
      }
    }

    switch(opstat){
    case MOUSE_OPSTATE.CTRL :
      if( (prev_mouse_Button & 1)!=0 && (mb & 1)==0 ){  // L-button on  => off
        prev_mouse_Button=mb ;
        opstat=MOUSE_OPSTATE.IDLE ;
        version(MOUSE_OP_TUNED){
          ShowCursor(true) ;
        }
      }else if( (prev_mouse_Button & 1)!=0 && (mb & 1)!=0 ){  // L-button on => on
        int x,y ;
        getRawMouseCurPos(&x, &y) ;
        version(MOUSE_OP_TUNED){
          dist_y+=(y-start_y) ;
        }else{
          dist_y=(y-start_y) ;
        }
        double rate=double(dist_y)/30.0 ;
        double dispscale=prev_scale * pow((1.0 + getZoomrate()*SGN(rate)), ABS(rate)) ;
        version(MOUSE_OP_TUNED){
          scaling(dispscale,start_x,start_y) ;
          //マウスカーソル位置をstart_x,y位置に戻すと何故か再描画かかからなかったので
          //RedrawWindow()で強制的に再描画を発生させている。なんかイマイチ。
          //タブレットの場合、カーソル位置が絶対位置の為、カーソル元の位置に戻すと
          //距離が大きくなり過ぎるのでダメ。
          POINT point ;
          point.x=start_x ;
          point.y=start_y ;
          ClientToScreen(getHWND(), &point) ;
          SetCursorPos(point.x,point.y) ;
          RedrawWindow(getHWND(),null,null,RDW_UPDATENOW) ;
        }else{
          scaling(dispscale,scale_center_x,scale_center_y) ;
        }
      }
      break ;
    case MOUSE_OPSTATE.NONE :
      if( (prev_mouse_Button & 1)!=0 && (mb & 1)==0 ){  // L-button on  => off
        prev_mouse_Button=mb ;
        opstat=MOUSE_OPSTATE.IDLE ;
      }else if( (prev_mouse_Button & 1)!=0 && (mb & 1)!=0 ){  // L-button on => on
        int x,y ;
        getRawMouseCurPos(&x, &y) ;
        setViewPosition(viewpos_x-(x-start_x),viewpos_y-(y-start_y)) ;
      }
      break ;
    default :
      break ;
    }

    return(1) ;
  }

  override int keyboard(int type, int acs_stat, int key, int mousex, int mousey){
    switch( type ){
    case 0:
      if( (acs_stat & CNTL_KEY)!=0 ){
        switch(key){
        case  6: /*C-f*/ scrolling( 1, 0,true) ; break ;
        case  2: /*C-b*/ scrolling(-1, 0,true) ; break ;
        case 14: /*C-n*/ scrolling( 0, 1,true) ; break ;
        case 16: /*C-p*/ scrolling( 0,-1,true) ; break ;
        case 22: /*C-v*/ scaling( 1.0,true) ; break ;
        case 26: /*C-z*/ scaling(-1.0,true) ; break ;
        default :
          return(0) ;
          break ;
        }
      }else{
        switch(key){
        case 'o': scaling(1.0) ; break ;
        case 'f': fitting( true, true) ; break ;
        case 'w': fitting( true,false) ; break ;
        case 'h': fitting(false, true) ; break ;
        case 'q':
          sendmsgtoparent(WM_CLOSE, 0,0) ;
          break ;
        default :
          return(0) ;
          break ;
        }
      }
      return(1) ;
      break ;

    case 1:
      switch(key){
      case VK_RIGHT : scrolling( 1, 0,true) ; break ;
      case VK_LEFT  : scrolling(-1, 0,true) ; break ;
      case VK_DOWN  : scrolling( 0, 1,true) ; break ;
      case VK_UP    : scrolling( 0,-1,true) ; break ;
      default : return(0) ;break ;
      }
      return(1) ;
      break ;

    default :
      return(0) ;
      break ;
    }
    return(0) ;
  }

  int scrolling(int dx, int dy, bool step=false){
    if( step==true ){
      int step_dx=getRollstep() * SGN(dx) ;
      int step_dy=getRollstep() * SGN(dy) ;
      getViewPosition(&viewpos_x, &viewpos_y) ;
      setViewPosition(viewpos_x+step_dx,viewpos_y+step_dy) ;
    }else{
      getViewPosition(&viewpos_x, &viewpos_y) ;
      setViewPosition(viewpos_x+dx,viewpos_y+dy) ;
    }
    return(0) ;
  }

  int scaling(double scale, bool step=false, bool wincenter=false){
    if( step==true ){
      double dispscale=getScale() ;
      if( dispscale>maxscale ) dispscale=maxscale ;
      else if( dispscale<minscale ) dispscale=minscale ;
      dispscale *= 1.0+getZoomrate()*SGN(scale) ;
      setScale(dispscale,true) ;
    }else{
      double dispscale=scale ;
      if( dispscale>maxscale ) dispscale=maxscale ;
      else if( dispscale<minscale ) dispscale=minscale ;
      setScale(dispscale,wincenter) ;
    }
    return(0) ;
  }

  int scaling(double scale, double px=0.0, double py=0.0){
    double dispscale=scale ;
    if( dispscale>maxscale ) dispscale=maxscale ;
    else if( dispscale<minscale ) dispscale=minscale ;
    setScale(dispscale,px,py) ;
    return(0) ;
  }

  int fitting(bool fitw=true, bool fith=true){
    for( int i=0 ; i<2 ; i++ ){
      double ww=getWindowWidth() ;
      double wh=getWindowHeight() ;
      double iw=getImageBufferWidth() ;
      double ih=getImageBufferHeight() ;
      if( fitw==true && fith==true ){
        if( ww>0.0 && wh>0.0 && iw>0.0 && ih>0.0 ){
          if( ww/iw*ih>=wh ){
            scaling(wh/ih) ;
          }else{
            scaling(ww/iw) ;
          }
        }
      }else if( fitw==true && fith==false ){
        if( ww>0.0 && iw>0.0 ){
          scaling(ww/iw) ;
        }
      }else if( fitw==false && fith==true ){
        if( wh>0.0 && ih>0.0 ){
          scaling(wh/ih) ;
        }
      }
    }
    return(0) ;
  }
}
