import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.Frame;
import win32frame.WindowMenu;
import win32frame.StatusBar;
import win32frame.MainWindow ;
import win32frame.FileDialog ;
import win32frame.MsgBox ;
import win32frame.FontSelDialog ;
import win32frame.Launcher;
import util.BitMap ;
import util.Image;

import MyButton;
import GraphicPanel;
import OpHelpDialog;

import std.stdio;
import std.string;
import std.utf;
import std.file;
import std.path;
import std.getopt;
import core.exception;

const string VERSIONSTRING="Ver0.01" ;

enum FILE{
  ROOT  =0x100,
  OPEN  =0x101,
  EXIT  =0x104
} ;
enum VIEW{
  ROOT  =0x300,
  FIT   =0x301,
  FITW  =0x302,
  FITH  =0x303,
  ORG   =0x304,
  ZMIN  =0x305,
  ZMOT  =0x306,
  AFIT  =0x307,
  ALPHA =0x308,
  HDTBOX=0x309
} ;
enum HELP{
  ROOT    =0x400,
  VERSION =0x401,
  OPHELP  =0x402
} ;

enum TOOLBOX{
  PREVFILE =0xa01,
  NEXTFILE =0xa02,
  TOGGLEFIT=0xa03,
  RELOAD   =0xa04
} ;

struct FileList
{
  string   curpath="" ;
  string[] namelist;
  size_t   cur=0 ;
  bool[string] extlist ;

  bool empty(){
    return( namelist.length==0 ? true : false ) ;
  }

  string getcurpath(){
    return(curpath) ;
  }

  string nextfile(){
    if( !empty() ){
      for( size_t i ; i<namelist.length ; i++ ){
        cur = (cur+1) % namelist.length ;
        string ext=toLower(extension(namelist[cur])) ;
        if( ext in extlist ) break ;
      }
      return(namelist[cur]) ;
    }
    return("") ;
  }

  string prevfile(){
    if( !empty() ){
      for( size_t i ; i<namelist.length ; i++ ){
        cur = (cur-1 + namelist.length) % namelist.length ;
        string ext=toLower(extension(namelist[cur])) ;
        if( ext in extlist ) break ;
      }
      return(namelist[cur]) ;
    }
    return("") ;
  }

  string curfile(){
    if( !empty() ){
      return(namelist[cur]) ;
    }
    return("") ;
  }

  int reload(string path, string fname){
    delete namelist ;
    namelist.length = 100 ;
    cur=0 ;
    int i;
    foreach( string name ; dirEntries(path, SpanMode.shallow)){
      if( isDir(name)==true ) continue ;
      if( i==namelist.length ){
        namelist.length=namelist.length*2 ;
      }
      namelist[i]=name ;
      if( baseName(name) == baseName(fname) ){
        cur=i ;
      }
      i++ ;
    }
    namelist.length=i ;
    Image img = new Image() ;
    extlist = img.getKnownExtlist() ;
    curpath=path ;
    return(0) ;
  }

  void debugprint(){
    foreach( i, name ; namelist ){
      if( i==cur ){
        writef("*") ;
      }
      writef("%d:%s\n",i,name) ;
    }
  }
}

struct ImageInfo
{
  string name ;
  int    width ;
  int    height ;
}

struct LauncherInfo
{
  int x,y;
  int w,h;
}

class myMainWindow : MainWindow {
  WindowMenu   menuroot  ;
  StatusBar    statusbar ;
  GraphicPanel gpanel ;
  MyButton     button0 ;
  MyButton     button1 ;
  MyButton     button2 ;
  MyButton     button3 ;
  Launcher     launcher ;
  LauncherInfo launcherinfo ;
  OpHelpDialog ophelpdlog ;
  string[]     args ;
  FileList     flist ;
  ImageInfo    imginfo ;
  bool         autofit=false ;
  bool         enalpha=true ;
  bool         noscrollbar=false ;
  bool         hidetoolbox=false ;
  bool         tbfit=false;

  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = LoadIcon(GetModuleHandle(null), toUTFz!(wchar*)("WAPICON"));
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name,ClassName) ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      if( ophelpdlog !is null ){
        ophelpdlog.setResult(0) ;
      }
      PostQuitMessage(0);
      break ;

    case WM_CREATE:
      menuroot = new WindowMenu() ;
      menuroot.create() ;
      menuroot.addMenu(             MFT_STRING, "File"      , FILE.ROOT ) ;
      menuroot.addItem( "File",     MFT_STRING, "Open"      , FILE.OPEN ) ;
      menuroot.addItem( "File",     MFT_STRING, "Exit"      , FILE.EXIT ) ;

      menuroot.addMenu(             MFT_STRING, "View"      , VIEW.ROOT ) ;
      menuroot.addItem( "View",     MFT_STRING, "Fit images to window size"   , VIEW.FIT  ) ;
      menuroot.addItem( "View",     MFT_STRING, "Fit images to window width"  , VIEW.FITW ) ;
      menuroot.addItem( "View",     MFT_STRING, "Fit images to window height" , VIEW.FITH ) ;
      menuroot.addItem( "View",     MFT_STRING, "Original size"               , VIEW.ORG  ) ;
      menuroot.addItem( "View",     MFT_STRING, "Zoom In"    , VIEW.ZMIN ) ;
      menuroot.addItem( "View",     MFT_STRING, "Zoom Out"   , VIEW.ZMOT ) ;
      menuroot.addItem( "View",     MFT_SEPARATOR, ""        , 0 ) ;
      menuroot.addItem( "View",     MFT_STRING, "Auto Fit images to window size" , VIEW.AFIT ) ;
      menuroot.addItem( "View",     MFT_STRING, "Blend Alpha", VIEW.ALPHA ) ;
      menuroot.addItem( "View",     MFT_STRING, "Hide ToolBox",VIEW.HDTBOX) ;

      menuroot.addMenu(             MFT_STRING, "Help"      , HELP.ROOT ) ;
      menuroot.addItem( "Help",     MFT_STRING, "Version"   , HELP.VERSION) ;
      menuroot.addItem( "Help",     MFT_STRING, "Operation" , HELP.OPHELP) ;

      setMenu(menuroot) ;

      statusbar = new StatusBar() ;
      statusbar.addStyle(CCS_BOTTOM | SBARS_SIZEGRIP) ;
      addStatusBar(statusbar, "") ;

      DragAcceptFiles(hwnd,true) ;

      launcherinfo.x=0 ;
      launcherinfo.y=0 ;
      launcherinfo.w=128 ;
      launcherinfo.h=32 ;
      launcher = new Launcher("Launcher") ;
      launcher.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      launcher.setPosition(launcherinfo.x,launcherinfo.y,10,10) ; //MoveWindow()でサイズ変更が伝わらない対策
      launcher.initGrid(4,1,32,32,true) ;
      addObject(launcher, cast(HMENU)(0xffff),false,false) ;

      BitMap btnimg ;

      btnimg= new BitMap() ;
      btnimg.readImage(getHInstance(), toUTFz!(wchar*)("TOOLBTN0"), MAKEINTRESOURCE_T!(256)) ;
      button0= new MyButton("＜") ;
      button0.setPosition(0,0,100,20) ;
      button0.setBitMap(btnimg) ;
      launcher.addObject(button0, cast(HMENU)(TOOLBOX.PREVFILE)) ;

      btnimg= new BitMap() ;
      btnimg.readImage(getHInstance(), toUTFz!(wchar*)("TOOLBTN1"), MAKEINTRESOURCE_T!(256)) ;
      button1= new MyButton("□") ;
      button1.setPosition(0,0,100,20) ;
      button1.setBitMap(btnimg) ;
      launcher.addObject(button1, cast(HMENU)(TOOLBOX.TOGGLEFIT)) ;

      btnimg= new BitMap() ;
      btnimg.readImage(getHInstance(), toUTFz!(wchar*)("TOOLBTN2"), MAKEINTRESOURCE_T!(256)) ;
      button2= new MyButton("＞") ;
      button2.setPosition(0,0,100,20) ;
      button2.setBitMap(btnimg) ;
      launcher.addObject(button2, cast(HMENU)(TOOLBOX.NEXTFILE)) ;

      btnimg= new BitMap() ;
      btnimg.readImage(getHInstance(), toUTFz!(wchar*)("TOOLBTN3"), MAKEINTRESOURCE_T!(256)) ;
      button3= new MyButton("○") ;
      button3.setPosition(0,0,100,20) ;
      button3.setBitMap(btnimg) ;
      launcher.addObject(button3, cast(HMENU)(TOOLBOX.RELOAD)) ;

      if( hidetoolbox==true ) launcher.hide() ;

      gpanel = new GraphicPanel("GraphicPanel","GraphicPanel") ;
      gpanel.setPosition(0,0,600,420) ;
      gpanel.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      gpanel.setCentering(true) ;
      gpanel.setAutoFocus(true) ;
      //gpanel.setDisableNoScroll(true) ;
      gpanel.setDisableScrollBar(noscrollbar) ;
      gpanel.setEnableBlendAlpha(enalpha) ;
      addObject( gpanel, null, true, true ) ;

      if( args.length>1 ){
        try{
          if( isDir(args[1])==true ){
            sendmsg(WM_COMMAND, FILE.OPEN, cast(LPARAM)((buildNormalizedPath(absolutePath(args[1])) ~ "\0").ptr)) ;
          }else{
            loadimagewindow(args[1]) ;
          }
        }catch(Exception e){
          auto msgbox = new MsgBox(null, e.msg, "FileOpenError", MB_OK ) ;
          msgbox.create() ;
        }
      }
      break ;

    case WM_DROPFILES:
      int    filenum = DragQueryFile( cast(HDROP)(wp), -1, null, 0) ;
      wchar  wfname[] ;
      string fnamelist[] ;

      fnamelist.length=filenum ;
      for( int i=0 ; i<filenum ; i++ ){
        int namelen=DragQueryFile( cast(HDROP)(wp), i, null, 0) ;
        wfname.length=namelen+1 ;
        DragQueryFile( cast(HDROP)(wp), i, wfname.ptr, namelen+1) ;
        wfname.length=wfname.length-1 ;
        fnamelist[i]=std.utf.toUTF8(wfname.idup) ;
      }
      DragFinish( cast(HDROP)(wp) ) ;
      if( isDir(fnamelist[0])==true ){
        sendmsg(WM_COMMAND, FILE.OPEN, cast(LPARAM)((fnamelist[0] ~ "\0").ptr)) ;
      }else{
        loadimagewindow(fnamelist[0]) ;
      }
      break ;

    case WM_CHAR:
      char key=LOBYTE(cast(ushort)(wp)) ;
      switch(key){
      case 'n' :
        loadimagewindow(flist.nextfile()) ;
        break ;
      case 'p' :
        loadimagewindow(flist.prevfile()) ;
        break ;
      case 'r' :
        loadimagewindow(flist.curfile()) ;
        break ;
      default :
        break ;
      }
      break ;

    case WM_KEYDOWN:
      char key=LOBYTE(cast(ushort)(wp)) ;
      switch(key){
      case VK_SPACE :
        loadimagewindow(flist.nextfile()) ;
        break ;
      case VK_BACK :
        loadimagewindow(flist.prevfile()) ;
        break ;
      case VK_TAB :
        toggleHideToolBox() ;
        break ;
      default :
        break ;
      }
      break ;

    case WM_SIZE :
      if( gpanel !is null ){
        if( autofit==true ) gpanel.fitting(true,true) ;
      }
      if( launcher !is null ){
        MoveWindow(launcher.getHWND(), winW/2-launcherinfo.w/2, winH-(cast(int)(2.4*launcherinfo.h)), launcherinfo.w,launcherinfo.h, true) ;
      }
      break ;

    case WM_SETFOCUS :
      SetFocus(gpanel.getHWND()) ;
      break ;

    case WM_INITMENU:
      menuroot.checkItem("View",VIEW.AFIT,   autofit ) ;
      menuroot.checkItem("View",VIEW.ALPHA,  enalpha ) ;
      menuroot.checkItem("View",VIEW.HDTBOX, hidetoolbox ) ;
      break ;

    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
      if( msg==WM_RBUTTONDOWN ) SetCapture( hwnd ) ;
      if( msg==WM_RBUTTONUP   ){
        ReleaseCapture() ;
        toggleHideToolBox() ;
      }
      break ;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case FILE.OPEN :
        FileDialog fdlog = new FileDialog() ;
        string     fname ;
        string     dirname="" ;
        if( lp!=0 ){
          char[] tmpbuf ;
          tmpbuf.length = 4096 ;
          for( int i=0 ; ; i++ ){
            if( i==tmpbuf.length ) tmpbuf.length = tmpbuf.length*2 ;
            if( (cast(char *)(lp))[i]==0 ){
              tmpbuf.length=i ;
              break ;
            }
            tmpbuf[i]=(cast(char *)(lp))[i] ;
          }
          dirname = std.utf.toUTF8(tmpbuf) ;
        }else{
          if( flist.empty()==false ){
            dirname=flist.getcurpath() ;
          }
        }
        fname=fdlog.getLoadFileName("ALL {*.*}\0*.*\0"
                                    "JPEG {*.JPG}\0*.jpg\0"
                                    "PNG {*.PNG}\0*.png\0"
                                    "TGA {*.TGA}\0*.tga\0"
                                    "BMP {*.BMP}\0*.bmp\0"
                                    "PBM {*.PBM}\0*.pbm\0"
                                    "PGM {*.PGM}\0*.pgm\0"
                                    "PPM {*.PPM}\0*.ppm\0",
                                    hwnd,OFN_HIDEREADONLY,
                                    dirname) ;
        loadimagewindow(fname) ;
        break ;

      case FILE.EXIT :
        sendmsg(WM_CLOSE, 0,0) ;
        break ;

      case VIEW.FIT :
        gpanel.fitting(true,true) ;
        break ;
      case VIEW.FITW :
        gpanel.fitting(true,false) ;
        break ;
      case VIEW.FITH :
        gpanel.fitting(false,true) ;
        break ;
      case VIEW.ORG :
        gpanel.scaling(1.0) ;
        break ;
      case VIEW.ZMIN :
        gpanel.scaling(1.0,true) ;
        break ;
      case VIEW.ZMOT :
        gpanel.scaling(-1.0,true) ;
        break ;
      case VIEW.AFIT :
        autofit = !autofit ;
        menuroot.checkItem("View",VIEW.AFIT, autofit) ;
        if( autofit==true ) gpanel.fitting(true,true) ;
        break ;
      case VIEW.ALPHA :
        enalpha = !enalpha ;
        menuroot.checkItem("View",VIEW.ALPHA, enalpha) ;
        gpanel.setEnableBlendAlpha(enalpha) ;
        break ;
      case VIEW.HDTBOX :
        toggleHideToolBox() ;
        menuroot.checkItem("View",VIEW.HDTBOX, hidetoolbox) ;
        break ;

      case HELP.VERSION :
        auto msgbox = new MsgBox(hwnd, "ImageView   " ~ VERSIONSTRING, "Version", MB_OK) ;
        msgbox.create() ;
        break ;
      case HELP.OPHELP :
        string[] helpstr = [ "---------Mouse Operation-------------------",
                             " ",
                             "       L-drag  : Scroll                    ",
                             "Ctrl + L-drag  : Zoom in/out               ",
                             "       R-click : Toggle show/hide ToolBox  ",
                             "       Wheel   : Zoom in/out               ",
                             " ",
                             "---------Keyboard Shortcuts----------------",
                             " ",
                             " 'f'        : Fit images to window size    ",
                             " 'w'        : Fit images to window width   ",
                             " 'h'        : Fit images to window height  ",
                             " 'o'        : Original size                ",
                             " 'r'        : Reload current file          ",
                             " 'n' or SPC : Next file in directory       ",
                             " 'p' or BS  : Previous file in directory   ",
                             " TAB        : Toggle show/hide ToolBox     ",
                             " ",
                             " 'Ctrl-f' or → : Scroll right             ",
                             " 'Ctrl-b' or ← : Scroll left              ",
                             " 'Ctrl-p' or ↑ : Scroll up                ",
                             " 'Ctrl-n' or ↓ : Scroll down              ",
                             " 'Ctrl-v'       : Zoom in                  ",
                             " 'Ctrl-z'       : Zoom out                 ",
                             ] ;
        RECT wrect ;
        GetWindowRect(getMainWinHWND(),&wrect) ;
        if( (ophelpdlog is null) ||
            (ophelpdlog !is null) && (ophelpdlog.getDlogState()==1) ){
          ophelpdlog = new OpHelpDialog("Operation Help","ophelp") ;
          ophelpdlog.setHelpString(helpstr) ;
          ophelpdlog.setModeless() ;
          ophelpdlog.Create( wrect.left+100,wrect.top+100,300,400,hwnd,null ) ;
        }
        break ;

      case TOOLBOX.NEXTFILE :
        loadimagewindow(flist.nextfile()) ;
        break ;
      case TOOLBOX.PREVFILE :
        loadimagewindow(flist.prevfile()) ;
        break ;
      case TOOLBOX.TOGGLEFIT :
        double scale=gpanel.getScale() ;
        if( scale!=1.0 ){
          gpanel.scaling(1.0) ;
        }else{
          gpanel.fitting(true,true) ;
        }
        break ;
      case TOOLBOX.RELOAD :
        loadimagewindow(flist.curfile()) ;
        break ;

      default :
        break ;
      }
      break;

    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int toggleHideToolBox(){
    hidetoolbox = !hidetoolbox ;
    if( hidetoolbox==true ) launcher.hide() ;
    else                    launcher.show() ;
    return(0) ;
  }

  int loadimagewindow(string fname){
    int ret=-1 ;
    if( fname.length!=0 ){
      if( LoadImage(fname)==0 ){
        printStatusbar( format("Open %s (%d x %d)",imginfo.name,imginfo.width,imginfo.height) ) ;
        ret=0 ;
      }else{
        printStatusbar( format("Open failed %s",fname) ) ;
      }
    }else{
      ret=-2 ;
    }
    return(ret) ;
  }

  int LoadImage(string fname){
    int   ret=-1 ;
    Image image = new Image() ;

    try{
      if( image.open(fname)==0 ){
        if( gpanel !is null ){
          gpanel.setImageSize(image.getWidth(),image.getHeight()) ;
          uint[][] imgbuf=gpanel.getImageBuffer() ;
          if( imgbuf==null ){
            ret=-2 ;
          }else{
            int imgw=image.getWidth()  ;
            int imgh=image.getHeight() ;
            for( size_t i=0 ; i<imgh ; i++ ){
              uint[] img = image.readline(1,true) ;
              imgbuf[i][] = img[0..$] ;
              delete img ;
            }
            gpanel.reDisplay() ;
            ret=0 ;
            imginfo.name  =fname ;
            imginfo.width =imgw ;
            imginfo.height=imgh ;
            if( autofit==true ) gpanel.fitting(true,true) ;
          }
        }
      }
    }catch( Exception e ){
      gpanel.setImageSize(0,0) ;
      ret=-3 ;
    }catch( OutOfMemoryError e ){
      gpanel.setImageSize(0,0) ;
      ret=-3 ;
    }
    image.close() ;
    destroy(image) ;

    if( isDir(fname)==true ){
      flist.reload(fname,fname) ;
    }else{
      flist.reload(dirName(fname),fname) ;
    }
    return(ret) ;
  }

  int setARGS(string[] args){
    bool opt_help=false ;
    getopt(args,
           std.getopt.config.caseSensitive,
           "autofit|A",     &autofit,
           "noscrollbar|N", &noscrollbar,
           "hidetoolbox|T", &hidetoolbox,
           "help|h",        &opt_help
           ) ;
    if( opt_help==true ){
      writef("Usage: imageview [OPTIONs] [FILE or DIRECTORY]\n"
             " command-line help\n"
             "  -A, --autofit     Enable Auto Fit images to window size.\n"
             "  -N, --noscrollbar No scrollbar mode.\n"
             "  -T, --hidetoolbox Hide toolbox at startup.\n"
             "  -h, --help        Print this help.\n"
             );
      return(-1) ;
    }
    this.args = args ;
    return(0) ;
  }

}
